package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.secretbox.SecretBox
import com.ionspin.kotlin.crypto.secretbox.crypto_secretbox_NONCEBYTES
import kotlinx.serialization.Serializable
import net.sergeych.bipack.BipackDecoder
import net.sergeych.bipack.BipackEncoder

/**
 * Symmetric key implements authenticated encrypting with random nonce and optional fill.
 * Random fill is normally used when cryptanalysis of the message size is a threat.
 *
 * Do not call this constructor directly, use [randomInt] or deserialize it.
 *
 * __Algorithms:__
 *
 * - Encryption: XSalsa20 stream cipher.
 * - Authentication: Poly1305 MAC
 */
@Serializable
class SymmetricKey(
    val keyBytes: UByteArray,
) : CipherKey {

    /**
     * @suppress
     * nonce + ciphered data serialization aid
     */
    @Serializable
    data class WithNonce(
        val cipherData: UByteArray,
        val nonce: UByteArray,
    )

    /**
     * @suppress
     * add some random data to the end of the plain message
     */
    @Serializable
    data class WithFill(
        val data: UByteArray,
        val safetyFill: UByteArray? = null,
    )

    override fun encryptWithNonce(plainData: UByteArray, nonce: UByteArray, randomFill: IntRange?): UByteArray {
        require(nonce.size == nonceByteLength)

        val fill = randomFill?.let {
            require(it.start >= 0)
            randomBytes(randomInt(it))
        }
        val filled = BipackEncoder.encode(WithFill(plainData, fill))
        return SecretBox.easy(filled.toUByteArray(), nonce, keyBytes)
    }

    override val nonceBytesLength: Int = nonceByteLength

    override val encryptingTag: UByteArray by lazy { blake2b2l(keyBytes) }
    override val decryptingTag: UByteArray get() = encryptingTag

    override fun decryptWithNonce(cipherData: UByteArray, nonce: UByteArray): UByteArray =
        protectDecryption {
            BipackDecoder.decode<WithFill>(SecretBox.openEasy(cipherData, nonce, keyBytes).toByteArray())
                .data
        }

    companion object {
        /**
         * Create a secure random symmetric key.
         */
        fun random() = SymmetricKey(SecretBox.keygen())

        val nonceByteLength = crypto_secretbox_NONCEBYTES
    }

}